﻿; ------------------------------------------------------------------------------------
; EProxy Server
; SOCKS5 Proxy nach RFC 1928 und Authentification nach RFC 1929
; ------------------------------------------------------------------------------------


; * Changelog V 1.01:
;   - Add/Remove - Client werden nun Asyncron verarbeitet und wurden geordnet.
;   - New: CurrentDownloadRate(), CurrentUploadRate() zur Speedanzeige
;   - BugFix: Transmitter konnte unter Umständen zu früh freigegeben werden
;   - BugFix: Transmitt - Status konnte dem Client zu früh erteilt werden
;   - BugFix: Hook_CheckAlive(), Gültigkeitbereich für Socket-Handles hinzugefügt
;   - CleanUp: Unnötige Strukturelemente ausgemistet


; ToDo:
; Hook in Mutex - Lock einschließen
; HT einbauen, Connection - Allowed Check
; Strukturen und Code besser kommentieren
; Userlisten + Passwörter, Filterfunktionen

; ------------------------------------------------------------------------------------
;- Internal
; ------------------------------------------------------------------------------------

CompilerIf #PB_Compiler_Thread = #False
  CompilerError "Enable Threadsafe"  
CompilerEndIf

; XIncludeFile "DebugHelper.pb"
XIncludeFile "SafeNetwork.pbi"
XIncludeFile "SArray.pbi"
XIncludeFile "LinkedList.pbi"
XIncludeFile "JobQueue_MultiThread_LE.pbi"
XIncludeFile "EProxy_Header.pbi"

EnableExplicit

; ------------------------------------------------------------------------------------
; Hilfsmacros
; ------------------------------------------------------------------------------------

Macro _EProxy_AddDownloadedBytes(_EProxy, _Bytes)
  LockMutex(_EProxy#\Statistic\TDBMutex)
  _EProxy#\Statistic\TotalDownloadedBytes + _Bytes
  UnlockMutex(_EProxy#\Statistic\TDBMutex)
EndMacro

Macro _EProxy_AddUploadedBytes(_EProxy, _Bytes)
  LockMutex(_EProxy#\Statistic\TUBMutex)
  _EProxy#\Statistic\TotalUploadedBytes + _Bytes
  UnlockMutex(_EProxy#\Statistic\TUBMutex)
EndMacro

Macro _EProxy_Connections(_EProxy, _Con)
  LockMutex(_EProxy#\Statistic\TCMutex)
  _EProxy#\Statistic\TotalConnections + _Con
  UnlockMutex(_EProxy#\Statistic\TCMutex)
EndMacro

; ------------------------------------------------------------------------------------
; Proceduren
; ------------------------------------------------------------------------------------

; ClientDaten laden
Procedure EProxy_GetClient(*EProxy.EProxy_Data, Client, Lock = #True)
  Protected *User.EProxy_UserData
  
  With *EProxy
    If Lock
      LockMutex(\ClientMutex)  
    EndIf
    *User = SArray_Get(\Clients, Client)
    If Lock
      UnlockMutex(\ClientMutex)  
    EndIf
  EndWith
  
  ProcedureReturn *User
EndProcedure

; Client sperren
Procedure EProxy_LockClient(*EProxy.EProxy_Data, Client)
  Protected *User.EProxy_UserData
  Protected RtVar, Mutex
  
  With *EProxy
    LockMutex(\ClientMutex)
    *User = EProxy_GetClient(*EProxy, Client, #False)
    If *User
      If Not *User\Removing; Kein Sperren mehr möglich wärend des entfernens!
        Mutex = *User\PMutex
        RtVar = (Mutex Or #Null)
      EndIf
    EndIf
    UnlockMutex(\ClientMutex)
    
    If Mutex
      LockMutex(Mutex)
    EndIf
    
    ProcedureReturn RtVar
  EndWith
EndProcedure

; Client Entsperren
Procedure EProxy_UnlockClient(*EProxy.EProxy_Data, Client)
  Protected *User.EProxy_UserData
  Protected RtVar, Mutex
  
  With *EProxy
    LockMutex(\ClientMutex)
    *User = EProxy_GetClient(*EProxy, Client, #False)
    If *User
      Mutex = *User\PMutex
      RtVar = (Mutex Or #Null)
    EndIf
    UnlockMutex(\ClientMutex)
    
    If Mutex
      UnlockMutex(Mutex)
    EndIf
    
    ProcedureReturn RtVar
  EndWith
EndProcedure

Procedure _EProxy_AddClient(*UserManager.EProxy_UserManager)
  Protected *EProxy.EProxy_Data, Client, *User.EProxy_UserData
  
  With *UserManager
    *EProxy = \EProxy
    Client = \Client
    *User = \User
  EndWith
  FreeMemory(*UserManager)
  
  With *EProxy
  ; User einbinden
    LockMutex(\ClientMutex)
    LL_Add(\ClientList) : *User\LLEntry = LL_GetEntry(\ClientList)
    LL_Set(\ClientList, Client)
    SArray_Set(\Clients, Client, *User)
    UnlockMutex(\ClientMutex)
    
    ; Statistik
    _EProxy_Connections(*EProxy, 1)
  EndWith
EndProcedure

; Client anlegen
Procedure EProxy_AddClient(*EProxy.EProxy_Data, Client)
  Protected *User.EProxy_UserData = AllocateMemory(SizeOf(EProxy_UserData))
  Protected *UserManager.EProxy_UserManager
  
  With *EProxy
    ; Default setzen
    *User\State = #Client_Step_Ver
    *User\SendSlot = JQ_Init(1) ; SendeSlot erstellen
    *User\PMutex = CreateMutex() ; Zugriffsmutex für UserDaten
    
    ; UserManager konfigurieren
    *UserManager = AllocateMemory(SizeOf(EProxy_UserManager))
    *UserManager\EProxy = *EProxy
    *UserManager\Client = Client
    *UserManager\User = *User
    
    ; In die Warteschlange stellen
    JQ_Add(\ClientManager, @_EProxy_AddClient(), *UserManager)
  EndWith
  
EndProcedure

Procedure _EProxy_RemoveClient(*UserManager.EProxy_UserManager)
  Protected DoubleDeath, i, *EProxy.EProxy_Data, Suicide, Client, *User.EProxy_UserData
  
  With *UserManager
    *EProxy = \EProxy
    Client = \Client
    Suicide = \Suicide
    *User = \User
  EndWith
  
  FreeMemory(*UserManager)
  
  With *EProxy
    ; Sorgt dafür das alle hier getroffenen Sicherheits-Checks auch greifen
    If Suicide
      For i = 1 To 5
        If \LastClient 
          Break  
        EndIf
        Delay(1)
      Next
    EndIf
      
    ; Dieser Lock kann erst ausgeführt werden wenn alle anderen Mutex - Locks (die zwischen LockClient() und
    ; *User\Removing = #True ja noch angefragt werden können) abgeabreitet sind. Erst nach diesem Lock ist 
    ; sichergestellt das im restlichen Source nicht mehr auf *User zugegriffen wird.
    LockMutex(*User\PMutex)
    UnlockMutex(*User\PMutex)
    FreeMutex(*User\PMutex)

    ; Empfänger stoppen
    If *User\Transmitter ; Gibt es einen Transmitter
      *User\Transmitter\Stop + 1 ; Stop setzen ...
      Repeat
        Delay(1)
      Until *User\Transmitter\Stop >= 2 ; Stop wurde bestätigt (>= ist wichtig bei Suizide)
      ; Beenden wurde vom Innenteil ausgelöst aber der Transmitter meldet Selbstmord? => Beide Verbindungen unterbrochen!
      FreeMemory(*User\Transmitter)
      *User\Transmitter = #Null
    EndIf 

      ; Sender stoppen
    JQ_WaitUntilFinish(*User\SendSlot) ; Warten bis alles gesendet wurde
    JQ_Deinit(*User\SendSlot) ; Slot beenden
      
    ; Beendet die Verbindung nach außen ...
    If *User\Connection And (Not Suicide)
      CloseNetworkConnection(*User\Connection)
    EndIf

    \LastClient = #Null
      
    ; Beendet die Verbindung nach innen ...
    If Suicide
      CloseNetworkConnection(Client)
    EndIf  
      
    ; Client austragen, aus Liste und SArray
    LockMutex(\ClientMutex)
    SArray_Set(\Clients, Client, #Null)
    LL_SetEntry(\ClientList, *User\LLEntry)
    LL_Delete(\ClientList)
    UnlockMutex(\ClientMutex)
    
    ; User freigeben ...
    FreeMemory(*User)
      
    ; Statistik
    _EProxy_Connections(*EProxy, -1)
  EndWith
    
EndProcedure

; Client löschen (so ziemlich die heikelste Stelle, hier kann viel schief gehen;
; 3 Threads mit unterschiedlichen geteilten Resourcen müssen hier geprüft und teilweise
; beendet werden)
Procedure EProxy_RemoveClient(*EProxy.EProxy_Data, Client, Suicide = #False)
  Protected *User.EProxy_UserData, isAlive, *UserManager.EProxy_UserManager
  
  With *EProxy
    If EProxy_LockClient(*EProxy, Client)
      *User = EProxy_GetClient(*EProxy, Client)
      If *User
        isAlive = (Not *User\Removing) ; Wird der User gerade entfernt?
        *User\Removing = #True ; Client kann nicht mehr gelockt werden
      EndIf
      EProxy_UnlockClient(*EProxy, Client) ; Client Unlocken
    ElseIf Not Suicide ; Der Client beendet sich wärend der Transmitter beendet
      \LastClient = Client
    EndIf
    
    If isAlive ; Gibt es den User noch?
      ; Usermanager konfigurieren
      *UserManager = AllocateMemory(SizeOf(EProxy_UserManager))
      *UserManager\EProxy = *EProxy
      *UserManager\Client = Client
      *UserManager\User = *User
      *UserManager\Suicide = Suicide
      
      ; In die Warteschlange stellen
      JQ_Add(\ClientManager, @_EProxy_RemoveClient(), *UserManager)
    EndIf

  EndWith
  
EndProcedure

; alle Client löschen
Procedure EProxy_RemoveAllClient(*EProxy.EProxy_Data)
  Protected Client, NextClient
  
  With *EProxy
    LockMutex(\ClientMutex)
    LL_Reset(\ClientList)
    NextClient = LL_Next(\ClientList)
    UnlockMutex(\ClientMutex)
    While NextClient
      LockMutex(\ClientMutex)
      Client = LL_Get(\ClientList) 
      UnlockMutex(\ClientMutex)
      
      EProxy_RemoveClient(*EProxy, Client)
      
      LockMutex(\ClientMutex)
      NextClient = LL_Next(\ClientList)
      UnlockMutex(\ClientMutex)
    Wend
  EndWith
  
EndProcedure

Declare EProxy_SendToClient(*EProxy.EProxy_Data, Client, Connection, *Mem, MemSize, ReSend = #False)

; An Client senden
Procedure EProxy_Sender(*Paket.EProxy_Sender)
  Protected RealSend
  Static Send
  
  With *Paket
    
    RealSend = SendNetworkData(\Connection, \Mem, \MemSize)
    
    If RealSend < \MemSize ; Was nicht gesendet werden konnten neu senden
      If RealSend = -1
        RealSend = #Null
         EProxy_SendToClient(\EProxy, \Client, \Connection, \Mem, \MemSize, #True)
      Else
        EProxy_SendToClient(\EProxy, \Client, \Connection, \Mem + RealSend, \MemSize - RealSend, #True)
      EndIf
    EndIf
    
      ; Statistik
    If \Client = \Connection
      _EProxy_AddDownloadedBytes(\EProxy, RealSend)
    Else
      _EProxy_AddUploadedBytes(\EProxy, RealSend)
    EndIf
  
    FreeMemory(\Mem)
    FreeMemory(*Paket)
  EndWith
EndProcedure
; Auf SendeSlot legen
Procedure EProxy_SendToClient(*EProxy.EProxy_Data, Client, Connection, *Mem, MemSize, ReSend = #False)
  Protected *Paket.EProxy_Sender
  Protected *User.EProxy_UserData
   
  If EProxy_LockClient(*EProxy, Client)  ; Zugriff muss gesichert werden
    *User = EProxy_GetClient(*EProxy, Client)
    
    ; Paket erstellen
    *Paket = AllocateMemory(SizeOf(EProxy_Sender))
    *Paket\EProxy = *EProxy
    *Paket\Client = Client
    *Paket\Connection = Connection
    *Paket\Mem = *Mem
    *Paket\MemSize = MemSize
    *Paket\Mem = AllocateMemory(MemSize)
    CopyMemory(*Mem, *Paket\Mem, MemSize)
    
    ; Auf SendeSlot legen
    If ReSend
      JQ_Add(*User\SendSlot, @EProxy_Sender(), *Paket, #Job_HighestPriority)
    Else
      JQ_Add(*User\SendSlot, @EProxy_Sender(), *Paket, #Job_HighPriority)
    EndIf
  
    EProxy_UnlockClient(*EProxy, Client)
  EndIf

EndProcedure

; Methode senden
Procedure EProxy_SendMethod(*EProxy, Client, Method)
  Protected RtVar, *Method.SOCKS_Ver = AllocateMemory(SizeOf(Word))
  
  With *Method
    \ver = #SOCKS_Version
    \Method = Method
  EndWith
  
  RtVar = EProxy_SendToClient(*EProxy, Client, Client, *Method, SizeOf(Word))
  FreeMemory(*Method)
  
  ProcedureReturn RtVar
EndProcedure

; Username + Passwort checken
Procedure EProxy_CheckUserData(*EProxy.EProxy_Data, UserName$, Password$)
  Protected Auth = #SOCKS_AUTH_Failure
  
  ; Später einführen
  
  Auth = #SOCKS_AUTH_Ok
  
  ProcedureReturn Auth
EndProcedure

; Reply senden
Procedure EProxy_SendReply(*EProxy, Client, Reply, Address = 0, Port = 0, atyp = #SOCKS_ATyp_IPv4)
  Protected RtVar, Repl.SOCKS_Reply, Repl4.SOCKS_V4_Ver
  
  If atyp = #SOCKS_ATYP_SOCKS4
    With Repl4 ; SOCKS4 benötigt einen anderen Rückgabetyp
      \ver = #Null
      \rsv = Reply
      \port = Port2Octed(Port)
      \ip = Address
    EndWith
    RtVar = EProxy_SendToClient(*EProxy, Client, Client, @Repl4, SizeOf(SOCKS_V4_Ver))
  Else  
    With Repl
      \ver = #SOCKS_Version
      \rsv = $00
      \rep = Reply
      \atyp = #SOCKS_ATyp_IPv4
      \addr = Address
      \port = Port2Octed(Port)
    EndWith
    RtVar = EProxy_SendToClient(*EProxy, Client, Client, @Repl, SizeOf(SOCKS_Reply))
  EndIf
  
  ProcedureReturn RtVar
EndProcedure

; Prüft ob die Verbindung erlaubt ist
Procedure EProxy_isConnectionAllowed(*EProxy.EProxy_Data, Client, Address$, Port)
  Protected RtVar = #False
  
  RtVar = #True
  
  ProcedureReturn RtVar
EndProcedure

Procedure EProxy_Transmitter(*Transmitter.EProxy_Transmitter)
  Protected Connection, Client, *EProxy.EProxy_Data, ReceiveBlockSize
  Protected BlockCount, *Mem, GesMemSize, Rest
  
  With *Transmitter
    *EProxy = \EProxy
    Client = \Client
    Connection = \Connection
  EndWith
    
  With *EProxy
    
    ; Standartwerte setzen
    ReceiveBlockSize = \TransmitterReceiveBlockSize
    BlockCount = #Null
    GesMemSize = #Null
    *Mem = #Null
    
    Repeat ; Daten einlesen
      *Mem = ReAllocateMemory(*Mem, (BlockCount + 1)*ReceiveBlockSize)
      Rest = ReceiveNetworkData(Connection, *Mem + BlockCount*ReceiveBlockSize, ReceiveBlockSize)
      GesMemSize + Rest
      BlockCount + 1
    Until Rest<>ReceiveBlockSize
  
    If GesMemSize>0 ; Auf wahre größe Resizen ...
      *Mem = ReAllocateMemory(*Mem, GesMemSize)
      
      ; Weiterleiten
      EProxy_SendToClient(*EProxy, Client, Client, *Mem, GesMemSize)
    EndIf
          
    ; Speicher freigeben
    FreeMemory(*Mem)
        
  EndWith
EndProcedure


; Transmitter
Procedure EProxy_PermanentTransmitter(*Transmitter.EProxy_Transmitter)
  Protected Connection, mClient, *EProxy.EProxy_Data, NCE, *User.EProxy_UserData
    
  With *Transmitter
    *EProxy = \EProxy
    mClient = \Client
    Connection = \Connection
  EndWith
  
  If EProxy_LockClient(*EProxy, mClient)
    *User = EProxy_GetClient(*EProxy, mClient)
    EProxy_UnlockClient(*EProxy, mClient)
  Else
    *Transmitter\SuicideDone = #True
    *Transmitter\Stop = 2
    EProxy_RemoveClient(*EProxy, mClient, #True)
    ProcedureReturn #True  
  EndIf
  
  With *EProxy   
    Repeat
      
      If *Transmitter\Stop ; StopSignal
        *Transmitter\Stop = 2 ; Bestätigen
        ProcedureReturn #True
      EndIf
      
      ; Client-Event abfragen
      NCE = NetworkClientEvent(Connection)
      
      ; Auf Meldungen warten ...
      If NCE = #PB_NetworkEvent_Data ; Daten laden und weiterleiten
        EProxy_Transmitter(*Transmitter)
        
      Else
        If NCE = #PB_NetworkEvent_Disconnect ; Server hat die Verbindung beendet ...
          ; Verbindung beenden
          *Transmitter\Stop = 2
          *Transmitter\SuicideDone = #True
          EProxy_RemoveClient(*EProxy, mClient, #True)
          ProcedureReturn #True  
        EndIf
        
        Delay(1) ; Einstellbar machen
      EndIf
      
    ForEver
  EndWith
EndProcedure

; Verbindung herstellen
Procedure EProxy_TryConnect(*EProxy.EProxy_Data, Client, Address$, Port, *Addr.SOCKS_Addr)
  Protected RtVar, Connection
  Protected *User.EProxy_UserData
  Protected *Transmitter.EProxy_Transmitter
  
  RtVar = #SOCKS_REP_GeneralSocksServerFailure
  
  If EProxy_LockClient(*EProxy, Client)  ; Zugriff muss gesichert werden
    *User = EProxy_GetClient(*EProxy, Client)
    With *EProxy
      If EProxy_isConnectionAllowed(*EProxy, Client, Address$, Port) ; Filer
      
        Connection = OpenNetworkConnection(Address$, Port)
    
        If Connection ; Verbindung OK, Transmitter erstellen
        
          *Addr\Port = Port
          *Addr\Addr = 16777343 ; 127.0.0.1
        
          ; Transmitter erstellen
          *Transmitter = AllocateMemory(SizeOf(EProxy_Transmitter))
              
          *Transmitter\EProxy = *EProxy
          *Transmitter\Client = Client
          *Transmitter\Connection = Connection
        
          ; Verbindung speichern ....
          *User\Connection = Connection
          *User\Transmitter = *Transmitter
        
          ; Transmitter launchen
          CreateThread(@EProxy_PermanentTransmitter(), *Transmitter)
          RtVar = #SOCKS_REP_Succeeded
         
        Else ; Nicht erreichbar, leider
          RtVar = #SOCKS_REP_HostUnreachable
        EndIf
      Else ; ProxyFilter hat zugeschlagen! Zensur ^^
        RtVar = #SOCKS_REP_ConnectionNotAllowedByRuleset
      EndIf
    
    EndWith
    
    EProxy_UnlockClient(*EProxy, Client)
  EndIf
  
  ProcedureReturn RtVar
EndProcedure

; CMD Auswerden
Procedure _EProxy_ProcessCMD(*CMDProcess.SOCKS_CMDProcess)
  Protected *EProxy.EProxy_Data, Client, *Mem, MemSize, *Ver4.SOCKS_V4_Ver
  Protected Target$, TargetSize, TargetPort, *Cmd.SOCKS_Cmd, Addr.SOCKS_Addr, Rep
  Protected User$, *User.EProxy_UserData
  
  With *CMDProcess
    *EProxy = \EProxy
    Client = \Client
    *Mem = \Mem
    MemSize = \MemSize
  EndWith
  
  FreeMemory(*CMDProcess)
  
  *Cmd = *Mem
  
  With *EProxy
 
    
    ; Überprüfung
    If MemSize>SizeOf(SOCKS_Cmd) And *Cmd\ver = #SOCKS_Version And *Cmd\rsv = $00; Inkorrekte Daten
      
      Select *Cmd\cmd ; Was soll gemacht werden
          
        Case #SOCKS_CMD_Connect ; Einfache TCP - Verbindung
          Select *Cmd\atyp
            Case #SOCKS_ATYP_IPv4 ; IPv4 Verbindung
              ; Ziel ermitteln                
              If MemSize = (SizeOf(SOCKS_Cmd) + SizeOf(Long) + SizeOf(Word)) ; Größe prüfen
                Target$ = IPString(PeekL(*Cmd + SizeOf(SOCKS_Cmd)))
                TargetPort = Octed2Port(PeekW(*Cmd + SizeOf(SOCKS_Cmd) + SizeOf(Long)))
                
                ; Verbindung herstellen :D
                Rep = EProxy_TryConnect(*EProxy, Client, Target$, TargetPort, @Addr)
              Else
                Rep = #SOCKS_REP_GeneralSocksServerFailure
              EndIf
                
            Case #SOCKS_ATYP_DomainName ; Verbindung zu Domainnamen
              ; Ziel ermitteln
              TargetSize = PeekA(*Cmd + SizeOf(SOCKS_Cmd))
             
              If MemSize = (SizeOf(SOCKS_Cmd) + SizeOf(Byte) + SizeOf(Word) + TargetSize) ; Größe prüfen
                Target$ = PeekS(*Cmd + SizeOf(SOCKS_Cmd) + SizeOf(Byte), TargetSize)
                TargetPort = Octed2Port(PeekW(*Cmd + SizeOf(SOCKS_Cmd) + SizeOf(Byte) + TargetSize))
                 
                ; Verbindung herstellen :D
                Rep = EProxy_TryConnect(*EProxy, Client, Target$, TargetPort, @Addr)
              Else
                Rep = #SOCKS_REP_GeneralSocksServerFailure
              EndIf
                
            Case #SOCKS_ATYP_SOCKS4 ; Socks4 Support
              ; Verbindung wurde über SOCKS4 angefordert
             If MemSize >= (SizeOf(SOCKS_Cmd) + SizeOf(SOCKS_V4_Ver) + SizeOf(Byte)) ; Mindestgröße prüfen
                *Ver4 = *Mem + SizeOf(SOCKS_CMD)
                Rep = EProxy_TryConnect(*EProxy, Client, IPString(*Ver4\ip), Octed2Port(*Ver4\port), @Addr)
                If Rep = #SOCKS_REP_Succeeded ; Rückgabe muss für SOCKS4 umgeschrieben werden
                  Rep = #SOCKS_V4_REP_Succeeded
                EndIf
              EndIf
                
            Case #SOCKS_ATYP_IPv6 ; IPv6 Verbindung
                ; Zurzeit kein IPv6
              Rep = #SOCKS_REP_AddressTypeNotSupported
              
          EndSelect
            
        Case #SOCKS_CMD_Bind ; Bind
          Rep = #SOCKS_REP_CommandNotSupported
            
        Case #SOCKS_CMD_UDPAssociate ; UDP
          Rep = #SOCKS_REP_CommandNotSupported
            
      EndSelect
    Else
      Rep = #SOCKS_REP_GeneralSocksServerFailure
    EndIf
    
    ; Reply senden
    EProxy_SendReply(*EProxy, Client, Rep, Addr\Addr, Addr\Port, *Cmd\atyp)
    
    If (Rep <> #SOCKS_REP_Succeeded) And (Rep <> #SOCKS_V4_REP_Succeeded); Fehler
      EProxy_RemoveClient(*EProxy, Client, #True)
    Else ; Alles Ok, jetztwird übertragen
      If EProxy_LockClient(*EProxy, Client)  ; Zugriff muss gesichert werden
        *User = EProxy_GetClient(*EProxy, Client)
        *User\State = #Client_Step_Transmit ; Jetzt wird übertragen
        EProxy_UnlockClient(*EProxy, Client)
      EndIf
    EndIf
    
  EndWith
  
  FreeMemory(*Mem)
 
EndProcedure

; Thread CMD Prozessor Launcher
Procedure EProxy_ProcessCMD(*EProxy.EProxy_Data, Client, *Mem, MemSize)
  Protected *Mem2, *CMD.SOCKS_CMDProcess = AllocateMemory(SizeOf(SOCKS_CMDProcess))
  
  *Mem2 = AllocateMemory(MemSize)
  CopyMemory(*Mem, *Mem2, MemSize)
  
  With *CMD
    \EProxy = *EProxy
    \Client = Client
    \Mem = *Mem2
    \MemSize = MemSize
  EndWith
  
  CreateThread(@_EProxy_ProcessCMD(), *CMD)
EndProcedure

; ProxyDaten verarbeiten
Procedure EProxy_Process(*EProxy.EProxy_Data, Client, *Mem, MemSize)
  Protected *User.EProxy_UserData
  Protected i, MethFound, Auth, *CMD.SOCKS_CMD, State
  Protected uLen, pLen, UserName$, Password$
  Protected *Ver.SOCKS_Ver, Connection, *MemV4
  Protected *Transmitter.EProxy_Transmitter
  
  With *EProxy
    
    If EProxy_LockClient(*EProxy, Client)  ; Zugriff muss gesichert werden
      *User = EProxy_GetClient(*EProxy, Client)
      State = *User\State
      EProxy_UnlockClient(*EProxy, Client)
    Else
      ProcedureReturn #False
    EndIf
    
    Select State ; Wo sind wir grade? ^^
      Case #Client_Step_Ver ; Anmeldung und festlegung der Methoden
        *Ver = *Mem
        If *Ver\ver = $04 ; SOCKSv4
          ; SOCKS4 Anfrage umbauen das der CMD - Prozessor damit umgehen kann
          ; CMD - Header vor die SOCKS - 4 Anfrage setzen
          *MemV4 = AllocateMemory(MemSize + SizeOf(SOCKS_CMD))
          CopyMemory(*Mem, *MemV4 + SizeOf(SOCKS_CMD), MemSize)
          MemSize + SizeOf(SOCKS_CMD)
          
          ; SOCKS5 Header schreiben
          *CMD = *MemV4
          *CMD\ver = #SOCKS_Version
          *CMD\cmd = #SOCKS_CMD_Connect
          *CMD\atyp = #SOCKS_ATYP_SOCKS4 ; Kleiner Trick, SOCKS4 wird als eigener Verbindungstyp definiert
          *Cmd\rsv = $00
          
          EProxy_ProcessCMD(*EProxy, Client, *MemV4, MemSize) ; Verarbeiten
          
          FreeMemory(*MemV4)
          
        ElseIf *Ver\ver = #SOCKS_Version ; Passt :D
          
          For i = 0 To *Ver\nMethods-1 ; Methode wählen
            Select *Ver\Methods[i]
              Case #SOCKS_Method_NoAuthenticationRequired
                If Not \NeedAuth ; Wir brauchen keine Auth
                  ; Jea, alles gut :D
                  EProxy_SendMethod(*EProxy, Client, #SOCKS_Method_NoAuthenticationRequired) ; Methode mitteilen
                  
                  If EProxy_LockClient(*EProxy, Client)  ; Zugriff muss gesichert werden
                    *User\State = #Client_Step_Cmd ; Status umschalten
                    EProxy_UnlockClient(*EProxy, Client)
                  EndIf
                  
                  ; Und weiter
                  MethFound = #True
                  Break
                  
                EndIf
                
              Case #SOCKS_Method_GSSAPI
                
                ; Erst implementieren
                ;
              Case #SOCKS_Method_UserNamePassword
                If \NeedAuth ; Wir brauchen keine Auth
                  EProxy_SendMethod(*EProxy, Client, #SOCKS_Method_UserNamePassword) ; Methode mitteilen
                  If EProxy_LockClient(*EProxy, Client)  ; Zugriff muss gesichert werden
                    *User\State = #Client_Step_Auth ; Status umschalten
                    EProxy_UnlockClient(*EProxy, Client)
                  EndIf
                Else
                  ; Jea, alles gut :D
                  EProxy_SendMethod(*EProxy, Client, #SOCKS_Method_NoAuthenticationRequired) ; Methode mitteilen
                  If EProxy_LockClient(*EProxy, Client)  ; Zugriff muss gesichert werden
                    *User\State = #Client_Step_Cmd ; Status umschalten
                    EProxy_UnlockClient(*EProxy, Client)
                  EndIf
                EndIf
                
                ; Und weiter
                MethFound = #True
                Break
                
            EndSelect
          Next
          
          If Not MethFound ; Keine passende Methode
            EProxy_SendMethod(*EProxy, Client, #SOCKS_Method_NoAcceptableMethod) ; Methode mitteilen
          EndIf
          
        EndIf
        
      Case #Client_Step_GSSAPI ; GSSAPI
        ; Noch implementieren
        
      Case #Client_Step_Auth ; Authentifikation
        ; Username und Passwort lesen.
        uLen = PeekA(*Mem + 1*SizeOf(Byte))
        UserName$ = PeekS(*Mem + 2*SizeOf(Byte), uLen)
        pLen = PeekA(*Mem + 2*SizeOf(Byte) + uLen)
        Password$ = PeekS(*Mem + 3*SizeOf(Byte) + uLen, pLen)
        
        ; Daten prüfen
        Auth = EProxy_CheckUserData(*EProxy, UserName$, Password$)
        EProxy_SendMethod(*EProxy, Client, Auth) ; Ergebniss mitteilen
        
        If Auth = #SOCKS_AUTH_Ok
          If EProxy_LockClient(*EProxy, Client)  ; Zugriff muss gesichert werden
            *User\State = #Client_Step_Cmd ; Status umschalten
            EProxy_UnlockClient(*EProxy, Client)
          EndIf
        EndIf
        
      Case #Client_Step_Cmd ; Commando - Ebene
        ; Thread Prozessor da ungültige Zieladdressen den Proxy hier ausbremsen könnten
        EProxy_ProcessCMD(*EProxy, Client, *Mem, MemSize)
         
    EndSelect
    
  EndWith
EndProcedure

; Empfänger
Procedure EProxy_Receiver(*EProxy.EProxy_Data)
  Protected Event, Client, EventServer, ReceiveBlockSize
  Protected *Mem, *pMem, Rest, GesMemSize, BlockCount, Connection
  Protected *User.EProxy_UserData
  
  With *EProxy
    Repeat
      If \Stop ; StopSignal
        SignalSemaphore(\Stop)
        ProcedureReturn #True
      EndIf
      
      Event = NetworkServerEvent()
      
      If Event ; Netzwerk Event da
        
        EventServer = EventServer()
        Client = EventClient()
        
        If EventServer = \Server
          Select Event
            Case #PB_NetworkEvent_Data

              ; Empfangsblockgröße für Service oder Transmit - Betrieb
              If EProxy_LockClient(*EProxy, Client)  ; Zugriff muss gesichert werden
                *User = EProxy_GetClient(*EProxy, Client)
                If *User\State = #Client_Step_Transmit ; Beim übertragen größere Blockgröße verwenden ...
                  ReceiveBlockSize = \TransmitterReceiveBlockSize
                  Connection = *User\Connection
                Else
                  ReceiveBlockSize = \ReceiveBlockSize
                  Connection = #Null
                EndIf
                EProxy_UnlockClient(*EProxy, Client)
              Else ; Der User wurde beendet ...
                Continue  
              EndIf
              
              BlockCount = #Null
              GesMemSize = #Null
              *Mem = #Null
              
              Repeat ; Daten einlesen
                *Mem = ReAllocateMemory(*Mem, (BlockCount + 1)*ReceiveBlockSize)
                Rest = ReceiveNetworkData(Client, *Mem + BlockCount*ReceiveBlockSize, ReceiveBlockSize)
                GesMemSize + Rest
                BlockCount + 1
              Until Rest<>ReceiveBlockSize
              
              If GesMemSize>0 ; auf wirkliche größe Resizen ...
                *Mem = ReAllocateMemory(*Mem, GesMemSize)
              EndIf
              
              If GesMemSize
                If Connection ; Weiterleiten oder Anfrage verarbeiten ?
                  EProxy_SendToClient(*EProxy, Client, Connection, *Mem, GesMemSize)
                Else
                  EProxy_Process(*EProxy, Client, *Mem, GesMemSize)
                EndIf
              EndIf
              
              If *Mem
                FreeMemory(*Mem)
              EndIf
              
            Case #PB_NetworkEvent_Connect
              EProxy_AddClient(*EProxy, Client)
              
            Case #PB_NetworkEvent_Disconnect
              EProxy_RemoveClient(*EProxy, Client)
              
          EndSelect
        EndIf
      Else
        Delay(1) ; CPU schonen...
      EndIf
    ForEver
  EndWith
EndProcedure


; ------------------------------------------------------------------------------------
;- Public
; ------------------------------------------------------------------------------------

; Proxy Initialisieren
ProcedureDLL EProxy_Init(Port = 1080)
  Protected Server, *EProxy.EProxy_Data
  
  ; Server erstellen
  Server = CreateNetworkServer(#PB_Any, Port)
  
  If Server
    ; Daten schreiben
    *EProxy = AllocateMemory(SizeOf(EProxy_Data))
    With *EProxy
      
      \ReceiveBlockSize = 64 ; Empfangspuffer für Proxy
      \TransmitterReceiveBlockSize = 64*1024 ; Empfangspuffer für Transmitter
      \Server = Server ; Server
      
      ; Statistik - Mutexen erstellen
      \Statistic\TDBMutex = CreateMutex()
      \Statistic\TUBMutex = CreateMutex()
      \Statistic\TCMutex = CreateMutex()
      
      ; Clientverwaltung erstellen
      \Clients = SArray_New()
      \ClientList = LL_New()
      \ClientMutex = CreateMutex()
      \ClientManager = JQ_Init(1)
      
      \Worker = CreateThread(@EProxy_Receiver(), *EProxy)
    EndWith
  EndIf
  
  ProcedureReturn *EProxy
EndProcedure

; Proxy beenden
ProcedureDLL EProxy_DeInit(*EProxy.EProxy_Data)
  With *EProxy
    
    ; Thread beenden
    \Stop = CreateSemaphore()
    WaitSemaphore(\Stop)
    FreeSemaphore(\Stop)
    
    ; Server beenden
    CloseNetworkServer(\Server)
    
    ; Clients löschen
    EProxy_RemoveAllClient(*EProxy)
    JQ_WaitUntilFinish(\ClientManager)
    JQ_DeInit(\ClientManager)
    
    ; Clientverzeichniss freigeben
    SArray_Free(\Clients)
    LL_Free(\ClientList)
    
    FreeMutex(\ClientMutex)
    FreeMutex(\Statistic\TDBMutex)
    FreeMutex(\Statistic\TUBMutex)
    FreeMutex(\Statistic\TCMutex)
    
    
    ; Rest freigeben
    FreeMemory(*EProxy)
  EndWith
EndProcedure

; Anzahl der Bytes die schon über den Proxy heruntergeladen wurden
Procedure.q EProxy_TotalDownloadedBytes(*EProxy.EProxy_Data)
  With *EProxy
    ProcedureReturn \Statistic\TotalDownloadedBytes
  EndWith
EndProcedure

; Anzahl der Bytes die schon über den Proxy hochgeladen wurden
Procedure.q EProxy_TotalUploadedBytes(*EProxy.EProxy_Data)
  With *EProxy
    ProcedureReturn \Statistic\TotalUploadedBytes
  EndWith
EndProcedure

; Aktuelle Downloadrate in Bytes/s
Procedure.d EProxy_CurrentDownloadRate(*EProxy.EProxy_Data)
  Protected Time, dTime, absTime.d, difBytes.d, Rate.d
  
  With *EProxy
    
    Time = ElapsedMilliseconds()
    dTime = Time - \Statistic\DownloadTimeStamp
    
    If Not \Statistic\DownloadTimeStamp
      \Statistic\DownloadTimeStamp = Time
    ElseIf \Statistic\TotalDownloadedBytes <> \Statistic\LastDownloadedBytes Or dTime < 2500
      
      If dTime > 1000 ; Alle Sekunden neu berechnen
        absTime.d = dTime/1000.0
        difBytes = \Statistic\TotalDownloadedBytes - \Statistic\LastDownloadedBytes
        \Statistic\DownloadRate = difBytes/absTime
        \Statistic\LastDownloadedBytes = \Statistic\TotalDownloadedBytes
        \Statistic\DownloadTimeStamp = Time
      EndIf
      
      Rate = \Statistic\DownloadRate
    EndIf
    
    ProcedureReturn Rate
  EndWith
EndProcedure

; Aktuelle Uploadrate in Bytes/s
Procedure.d EProxy_CurrentUploadRate(*EProxy.EProxy_Data)
  Protected Time, dTime, absTime.d, difBytes.d, Rate.d
  
  With *EProxy
    
    Time = ElapsedMilliseconds()
    dTime = Time - \Statistic\UploadTimeStamp
    
    If Not \Statistic\UploadTimeStamp
      \Statistic\UploadTimeStamp = Time
    ElseIf \Statistic\TotalUploadedBytes <> \Statistic\LastUploadedBytes Or dTime < 2500
      
      If dTime > 1000 ; Alle Sekunden neu berechnen
        absTime.d = dTime/1000.0
        difBytes = \Statistic\TotalUploadedBytes - \Statistic\LastUploadedBytes
        \Statistic\UploadRate = difBytes/absTime
        \Statistic\LastUploadedBytes = \Statistic\TotalUploadedBytes
        \Statistic\UploadTimeStamp = Time
      EndIf
      
      Rate = \Statistic\UploadRate
    EndIf
    
    ProcedureReturn Rate
  EndWith
EndProcedure

; Anzahl der Verbindungen die am Proxy aktiv sind
Procedure.q EProxy_TotalConnections(*EProxy.EProxy_Data)
  With *EProxy
    ProcedureReturn \Statistic\TotalConnections
  EndWith
EndProcedure
; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 209
; FirstLine = 170
; Folding = ------
; EnableThread
; EnableXP
; Executable = C:\Users\Alexander\Desktop\Proxy.exe
; CPU = 1
; EnablePurifier
; EnableCompileCount = 144
; EnableBuildCount = 2
; EnableExeConstant